"""
ref: https://github.com/ethereum/research/blob/711bd9532b4534ef5ae6277bd7afe625195506d5/zksnark/bn128_field_elements.py
"""
import klefki.const as const
from klefki.algebra.fields import FiniteField
from klefki.algebra.groups import EllipticCurveGroup
from klefki.algebra.groups import EllipicCyclicSubgroup
from klefki.curves.arith import short_weierstrass_form_curve_addition2


class FiniteFieldBN254(FiniteField):
    P = const.BN254_P


class EllipticCurveBN254(EllipticCurveGroup):
    """
    Twisted Edwards Form (standard)
    y^2 = x^3 + Ax^2 + x
    Montgomery Form
    By^2 = x^3 + A x^2 + x
    """
    A = const.BN254_A
    B = const.BN254_B
    N = const.BN254_N

    def op(self, g):
        if g == self.zero():
            return self
        if self == self.zero():
            return g
        field = self.id[0].type

        # a1,a3,a2,a4,a6 = 0, 0, 0, a, b
        x, y = short_weierstrass_form_curve_addition2(
            self.x, self.y,
            g.x, g.y,
            field(0),
            field(0),
            field(0),
            field(self.A),
            field(self.B),
            field
        )
        if x == y == field(0):
            return self.__class__(0)
        return self.__class__((x, y))


EllipticCurveBN254.G = EllipticCurveBN254((
    FiniteFieldBN254(const.BN254_Gx),
    FiniteFieldBN254(const.BN254_Gy)
))



from abc import abstractproperty
from .abstract import Field
from klefki.algorithms import extended_euclidean_algorithm
from klefki.algorithms import complex_truediv_algorithm
from klefki.algorithms import deg, poly_rounded_div


class FiniteField(Field):

    P = abstractproperty()

    def fmap(self, o):
        value = getattr(o, 'value', o)
        if isinstance(value, complex):
            return complex((value.real % self.P), (value.imag % self.P))
        return value % self.P

    def inverse(self):
        return self.__class__(self.P - self.value)

    def mod(self, a, b):
        if isinstance(a, complex):
            return complex((a.real % b), (a.imag % b))
        return a % b

    def sec_inverse(self):
        if isinstance(self.value, complex):
            return complex_truediv_algorithm(complex(1), self.value, self.__class__)

        gcd, x, y = extended_euclidean_algorithm(self.value, self.P)
        assert (self.value * x + self.P * y) == gcd
        # if gcd != 1:
        #     # Either n is 0, or p is not prime number
        #     raise ValueError(
        #         '{} has no multiplicate inverse '
        #         'modulo {}'.format(self.value, self.P)
        #     )
        return self.__class__(x % self.P)

    def op(self, g):
        if isinstance(g, int):
            g = self.type(g)
        return self.__class__(
            self.mod(
                (self.value + g.value), self.P
            )
        )

    def sec_op(self, g):
        if isinstance(g, int):
            g = self.type(g)
        return self.__class__(
            self.mod(
                (self.value * g.value), self.P
            )
        )


class PolyExtField(Field):
    # field
    F = abstractproperty()
    E = abstractproperty()

    def fmap(self, value):
        # map Maybe<int> top Field
        if value == 0:
            return self.zero().id
        if value == 1:
            return self.one().id
        if isinstance(value, list):
            assert len(value) == len(self.E)
            return [self.F(p) for p in value]
        # support inner field
        return self.F(value)

    def op(self, rhs):
        if isinstance(rhs, int):
            rhs = self.F(rhs)
        if isinstance(rhs, self.F):
            return self.type([x + rhs for x in self.id])
        return self.type([x + y for x, y in zip(self.id, rhs.id)])

    @classmethod
    def sec_identity(cls):
        return cls([cls.F.one()] + [cls.F.zero()] * (len(cls.E) - 1))

    @classmethod
    def identity(cls):
        return cls([cls.F.zero()] * len(cls.E))

    def inverse(self):
        return self.type([-x for x in self.id])

    def sec_inverse(self):
        field = self.F
        lm, hm = [self.F.one()] + [self.F.zero()] * \
            len(self.E), [self.F.zero()] * (len(self.E) + 1)
        low, high = self.id + [self.F.zero()], [self.F(m)
                                                for m in self.E] + [field(1)]
        while deg(low):
            r = poly_rounded_div(high, low, field)
            r += [field.zero()] * (len(self.E) + 1 - len(r))
            nm = [field(x) for x in hm]
            new = [field(x) for x in high]
            assert len(lm) == len(hm) == len(low) == len(
                high) == len(nm) == len(new) == len(self.E) + 1
            # xt Euclidean alog.
            for i in range(len(self.E) + 1):
                for j in range(len(self.E) + 1 - i):
                    nm[i+j] -= lm[i] * r[j]
                    new[i+j] -= low[i] * r[j]
            lm, low, hm, high = nm, new, lm, low
        return self.type([i / field(low[0]) for i in lm[:len(self.E)]])

    def sec_op(self, rhs):
        if isinstance(rhs, int):
            rhs = self.F(rhs)
        if isinstance(rhs, self.F):
            return self.__class__([c * rhs for c in self.value])
        else:
            b = [self.F.zero() for i in range(len(self.E) * 2 - 1)]
            inner_enumerate = list(enumerate(rhs.value))

            for i, eli in enumerate(self.value):
                for j, elj in inner_enumerate:
                    b[i + j] += eli * elj
            for exp in range(len(self.E) - 2, -1, -1):
                top = b.pop()
                for i, c in [(i, c) for i, c in enumerate(self.E) if c]:
                    b[exp + i] -= top * c
            return self.__class__([x for x in b])


PrimeField = FiniteField
Fq = FiniteField



"""
ref: https://github.com/ethereum/research/blob/711bd9532b4534ef5ae6277bd7afe625195506d5/zksnark/bn128_field_elements.py
"""
import klefki.const as const
from klefki.algebra.fields import FiniteField
from klefki.algebra.fields import PolyExtField
from klefki.algebra.groups import EllipticCurveGroup
from klefki.algebra.groups import EllipicCyclicSubgroup
from klefki.curves.arith import short_weierstrass_form_curve_addition2


class BN128FP(FiniteField):
    P = const.BN128_P


class BN128FP2(PolyExtField):
    F = BN128FP
    E = const.BN128_FP2_E


class BN128FP12(PolyExtField):
    F = BN128FP
    E = const.BN128_FP12_E

    @classmethod
    def from_fp(cls, v):
        return cls([v] + ([cls.F.zero()] * 11))


class ECGBN128(EllipticCurveGroup):
    A = const.BN128_A

    N = const.BN128_N

    def op(self, g):
        if g == self.zero():
            return self
        if self == self.zero():
            return g
        field = self.id[0].type

        # a1,a3,a2,a4,a6 = 0, 0, 0, a, b
        x, y = short_weierstrass_form_curve_addition2(
            self.x, self.y,
            g.x, g.y,
            field.zero(),
            field.zero(),
            field.zero(),
            field(self.A),
            field(self.B),
            field
        )
        if x == y == field.zero():
            return self.zero()
        return self.__class__((x, y))

    def twist(self):
        x, y = self.x, self.y
        if self == self.zero():
            return self.zero()
        if isinstance(x, BN128FP12) and isinstance(y, BN128FP12):
            return self
        elif isinstance(x, BN128FP2) and isinstance(y, BN128FP2):
            return self.twist_FP2_to_FP12(x, y)
        elif isinstance(x, BN128FP) and isinstance(y, BN128FP):
            return self.twist_FP_to_FP12(x, y)
        else:
            raise Exception("cannot twist curve to fp12")

    @classmethod
    def twist_FP_to_FP12(cls, x, y):
        assert isinstance(x, BN128FP)
        assert isinstance(y, BN128FP)
        ret = cls(BN128FP12.from_fp(x), BN128FP12.from_fp(y))
        assert ret.is_on_curve()
        return ret

    @classmethod
    def twist_FP2_to_FP12(cls, x, y):
        # "Twist" a point in E(FQ2) into a point in E(FQ12)
        zero = BN128FP.zero()
        one = BN128FP.one()
        w = BN128FP12([zero, one] + [zero] * 10)
        assert isinstance(x, BN128FP2)
        assert isinstance(y, BN128FP2)
        nx = BN128FP12([x.id[0]] + [zero] * 5 + [x.id[1]] + [zero] * 5)
        ny = BN128FP12([y.id[0]] + [zero] * 5 + [y.id[1]] + [zero] * 5)
        ret = cls((nx * w ** 2, ny * w**3))
        assert ret.is_on_curve()
        return ret

    @staticmethod
    def linefunc(P1, P2, T):

        # https://github.com/ethereum/research/blob/9a7b6825b0dee7a59a03f8ca1d1ec3ae7fb6d598/zksnark/bn128_curve.py
        assert P1 and P2 and T  # No points-at-infinity allowed, sorry
        x1, y1 = P1.x, P1.y
        x2, y2 = P2.x, P2.y
        xt, yt = T.x, T.y
        if x1 != x2:
            m = (y2 - y1) / (x2 - x1)
            return m * (xt - x1) - (yt - y1)
        elif y1 == y2:
            m = (x1**2 * 3) / (y1 * 2)
            return m * (xt - x1) - (yt - y1)
        else:
            return xt - x1

    @classmethod
    def miller_loop(cls, Q, P):
        # https://crypto.stanford.edu/pbc/notes/ep/miller.htm
        # ref: https://github.com/ethereum/research/blob/9a7b6825b0dee7a59a03f8ca1d1ec3ae7fb6d598/zksnark/bn128_pairing.py
        log_ate_loop_count = 63
        ate_loop_count = 29793968203157093288

        if Q == cls.zero() or P == cls.zero():
            return cls.one()

        R = Q
        f = BN128FP12.one()
        for i in range(log_ate_loop_count, -1, -1):
            f = f * f * cls.linefunc(R, R, P)
            R = R @ 2
            if ate_loop_count & (2**i):
                f = f * cls.linefunc(R, Q, P)
                R = R + Q
        Q1 = cls(Q.x ** BN128FP.P, Q.y ** BN128FP.P)
        nQ2 = cls(Q1.x ** BN128FP.P, (-Q1.y) ** BN128FP.P)
        f = f * cls.linefunc(R, Q1, P)
        R = R + Q1
        f = f * cls.linefunc(R, nQ2, P)
        R = R + nQ2
        return f ** ((BN128FP.P ** 12 - 1) // cls.N)

    @classmethod
    def pairing(cls, Q, P):
        """
        e(P, Q + R) = e(P, Qj * e(P, R)
        e(P + Q, R) = e(P, R) * e(Q, R)
        """
        return cls.miller_loop(Q.twist(), P.twist())

    @classmethod
    def e(cls, P, Q):
        return cls.pairing(P, Q)

    def is_on_curve(self):
        return self.y**2 - self.x**3 == self.B

    @property
    def B(self):
        if isinstance(self.x, BN128FP2):
            return BN128FP2([3, 0]) / BN128FP2([0, 1])
        elif isinstance(self.x, BN128FP12):
            return BN128FP12([3] + [0] * 11)
        return BN128FP(3)


ECGBN128.G1 = ECGBN128(
    BN128FP(const.BN128_G1x),
    BN128FP(const.BN128_G1y)
)

ECGBN128.G2 = ECGBN128(
    BN128FP2(const.BN128_G2x),
    BN128FP2(const.BN128_G2y)
)

ECGBN128.G = ECGBN128.G1



"""
from https://github.com/ethereum/research/blob/master/zksnark/code_to_r1cs.py
ref: https://medium.com/@VitalikButerin/quadratic-arithmetic-programs-from-zero-to-hero-f6d558cea649
"""

import ast
import inspect
from klefki.zkp.flatcode import Flattener
import types


# Adds a variable or number into one of the vectors; if it's a variable
# then the slot associated with that variable is set to 1, and if it's
# a number then the slot associated with 1 gets set to that number
def insert_var(arr, varz, var, used, reverse=False):
    if isinstance(var, str):
        if var not in used:
            raise Exception("Using a variable before it is set!")
        arr[varz.index(var)] += (-1 if reverse else 1)
    elif isinstance(var, int):
        arr[0] += var * (-1 if reverse else 1)

# Maps input, output and intermediate variables to indices


def get_var_placement(inputs, flatcode):
    return ['~one'] + [x for x in inputs] + ['~out'] + [
        c[1] for c in flatcode if c[1] not in inputs and c[1] != '~out']


# Convert the flattened code generated above into a rank-1 constraint system
def flatcode_to_r1cs(inputs, flatcode, field=int):
    # f.var
    varz = get_var_placement(inputs, flatcode)
    A, B, C = [], [], []
    used = {i: True for i in inputs}
    for x in flatcode:
        a, b, c = [field(0)] * len(varz), [field(0)] * \
            len(varz), [field(0)] * len(varz)
        if x[1] in used:
            raise Exception("Variable already used: %r" % x[1])
        used[x[1]] = True
        if x[0] == 'set':
            a[varz.index(x[1])] += field(1)
            insert_var(a, varz, x[2], used, reverse=True)
            b[0] = field(1)
        elif x[0] == '+' or x[0] == '-':
            c[varz.index(x[1])] = field(1)
            insert_var(a, varz, x[2], used)
            insert_var(a, varz, x[3], used, reverse=(x[0] == '-'))
            b[0] = field(1)
        elif x[0] == '*':
            c[varz.index(x[1])] = field(1)
            insert_var(a, varz, x[2], used)
            insert_var(b, varz, x[3], used)
        elif x[0] == '/':
            insert_var(c, varz, x[2], used)
            a[varz.index(x[1])] = field(1)
            insert_var(b, varz, x[3], used)
        A.append(a)
        B.append(b)
        C.append(c)
    return A, B, C

# Get a variable or number given an existing input vector


def grab_var(varz, assignment, var):
    if isinstance(var, str):
        return assignment[varz.index(var)]
    elif isinstance(var, int):
        return var
    else:
        raise Exception("What kind of expression is this? %r" % var)

# Goes through flattened code and completes the input vector


def assign_variables(inputs, input_vars, flatcode, field):
    varz = get_var_placement(inputs, flatcode)

    assignment = [field(0)] * len(varz)
    assignment[0] = field(1)
    for i, inp in enumerate(input_vars):
        assignment[i + 1] = field(inp)
    for x in flatcode:
        assignment[varz.index(x[1])] = {
            "set": lambda: field(grab_var(varz, assignment, x[2])),
            "+": lambda: field(grab_var(varz, assignment, x[2])) + field(grab_var(varz, assignment, x[3])),
            "-": lambda: field(grab_var(varz, assignment, x[2])) - field(grab_var(varz, assignment, x[3])),
            "*": lambda: field(grab_var(varz, assignment, x[2])) * field(grab_var(varz, assignment, x[3])),
            "/": lambda: field(grab_var(varz, assignment, x[2])) / field(grab_var(varz, assignment, x[3]))
        }[x[0]]()
    return assignment


def code_to_r1cs_with_inputs(code, input_vars, field):
    flatten = Flattener(code)
    inputs = flatten.inputs
    flatcode = flatten.flatten_code
    A, B, C = flatcode_to_r1cs(inputs, flatcode, field)
    r = assign_variables(inputs, input_vars, flatcode, field)
    return r, A, B, C


def mul(a, b):
    return list(map(lambda x: x[0] * x[1], zip(a, b)))


class R1CS:

    @staticmethod
    def parse(code, input_vals, field=int):
        s, A, B, C = code_to_r1cs_with_inputs(code, input_vals, field)
        return (s, A, B, C)

    @staticmethod
    def verify(s, A, B, C):
        ret = True
        for i in range(len(A)):
            ret = ret and sum(mul(A[i], s)) * \
                sum(mul(B[i], s)) == sum(mul(C[i], s))
        return ret

    @staticmethod
    def r1cs(fn_or_field=int, ctx={}):
        def wrapper(f):
            src = inspect.getsource(f)
            flatten = Flattener(src, ctx)
            f.inputs = flatten.inputs
            f.flatcode = flatten.flatten_code
            f.r1cs = flatcode_to_r1cs(f.inputs, f.flatcode, field)
            f.A = f.r1cs[0]
            f.B = f.r1cs[1]
            f.C = f.r1cs[2]
            f.var = get_var_placement(f.inputs, f.flatcode)
            f.src = src

            def wit(*args):
                return assign_variables(f.inputs, args, f.flatcode, field)
            f.witness = wit
            return f
        if isinstance(fn_or_field, types.FunctionType):
            f = fn_or_field
            field = int
            return wrapper(f)
            return ret
        else:
            field = fn_or_field
            return wrapper



"""
"""
from klefki.algebra.fields import FiniteField


def short_weierstrass_form_curve_addition2(
        u1, v1, u2, v2, a1, a3, a2, a4, a6, f: FiniteField
) -> (FiniteField, FiniteField):
    """
    https://core.ac.uk/download/pdf/10898289.pdf
    E_{W,a1a3a2a4a6} = v^2 + a1uv + a3v = u^3 + a2u^2 + a4u + a6
    for k256 curve: v^2 = u^3 + a * u + b
    a1,a3,a2,a4,a6 = 0, 0, 0, a, b
    for baby_jubjub curve: Bv^2 = x^3 + Ax^2 + x
    a1,a3,a2,a4,a6 = 0, 0, a, 1, 0
    """
    if f(u1) == f.zero() and f(v1) == f.zero():
        return (u2, v2)
    elif f(u2) == f.zero() and f(v2) == f.zero():
        return (u1, v1)
    elif u1 == u2:
        if v1 != v2:
            return (f.zero(), f.zero())
        elif v1*2 + u1 * a1 + a3 == f.zero():
            return (f.zero(), f.zero())
        else:
            lam = ((u1 ** 2) * 3 + u1 * a2 * 2 -
                   v1 * a1 + a4) / (v1 * 2 + u1 * a1 + a3)
            u3 = lam ** 2 + lam * a1 - a2 - u2 * 2
            v3 = lam * (u1 - u3) - v1 - a1 * u3 - a3
            return (u3, v3)
    else:
        lam = (v1 - v2) / (u1-u2)
        u3 = lam ** 2 + a1 * lam - a2 - u1 - u2
        v3 = lam * (u1-u3) - v1 - a1 * u3 - a3
        return (u3, v3)



from abc import ABCMeta, abstractmethod, abstractproperty
from klefki.algorithms import double_and_add_algorithm
from klefki.algorithms import complex_truediv_algorithm
from klefki.algorithms import fast_pow
from klefki.numbers import modular_sqrt
import sys


class Generic(metaclass=ABCMeta):

    __slots__ = ['value']

    @classmethod
    def construct(cls, name, **kwargs):
        return type(name, (cls, ), dict(**kwargs))

    def __init__(self, *args):
        if len(args) == 1:
            args = args[0]
            if isinstance(args, self.type):
                self.value = args.value
            else:
                self.value = self.fmap(args)
        else:
            self.value = self.fmap(args)

    def fmap(self, o):
        return o

    @property
    def type(self):
        return self.__class__

    @property
    def id(self):
        return self.value


class Groupoid(Generic):

    __slots__ = ()

    @abstractmethod
    def op(self, g: 'Group') -> 'Group':
        pass

    def __eq__(self, b) -> bool:
        return self.id == b.id

    def __lt__(self, b) -> bool:
        return self.id < b.id

    def __le__(self, b) -> bool:
        return self.id <= b.id

    def __gt__(self, b) -> bool:
        return self.id > b.id

    def __ge__(self, b) -> bool:
        return self.id >= b.id

    def __add__(self, g: 'Group') -> 'Group':
        '''
        Allowing call associativity operator via A + B
        Strict limit arg `g` and ret `res` should be subtype of Group,
        For obeying axiom `closure` (1)
        '''
        res = self.op(g)
        assert isinstance(res, type(self))
        return res

    def __radd__(self, g):
        return self.__add__(g)

    def __mul__(self, g: 'Group') -> 'Group':
        return self.__add__(g)

    def __repr__(self):
        return "%s::%s" % (
            type(self).__name__,
            self.id
        )

    def __str__(self):
        return str(self.id)


class SemiGroup(Groupoid):

    __slots__ = ()

    @abstractmethod
    def op(self, g: 'Group') -> 'Group':
        '''
        The Operator for obeying axiom `associativity` (2)
        '''
        pass


class Monoid(SemiGroup):
    __slots__ = ()

    @classmethod
    def zero(cls):
        return cls.identity()

    @classmethod
    def identity(cls):
        '''
        The value for obeying axiom `identity` (3)
        '''
        return cls(0)

    def __not__(self):
        return self is not self.identity()

    def scalar(self, times):
        while getattr(times, 'value', None) != None:
            times = times.id
        if times == 0:
            return self.identity()
        return double_and_add_algorithm(times, self, self.identity())

    def __matmul__(self, times):
        return self.scalar(times)

    def __pow__(self, times) -> 'Group':
        return self.scalar(times)

    def __xor__(self, times) -> 'Group':
        return self.__matmul__(times)


class Group(Monoid):
    __slots__ = ()

    @abstractmethod
    def inverse(self: 'Group') -> 'Group':
        '''
        Implement for axiom `inverse`
        '''
        pass

    def __sub__(self, g: 'Group') -> 'Group':
        '''
        Allow to reverse op via a - b
        '''
        return self.op(g.inverse())

    def __neg__(self) -> 'Group':
        return self.inverse()


class Ring(Group):
    __slots__ = ()

    @abstractmethod
    def sec_op(self, g: 'Field') -> 'Field':
        '''
        The Operator for obeying axiom `associativity` (2)
        '''
        pass

    def __mul__(self, g: 'Field') -> 'Field':
        '''
        Allowing call associativity operator via A * B
        Strict limit arg `g` and ret `res` should be subtype of Group,
        For obeying axiom `closure` (1)
        '''
        res = self.sec_op(g)
        assert isinstance(res, type(self)), 'result shuould be %s' % type(self)
        return res

    def __pow__(self, b, m=None):
        if hasattr(b, "value"):
            b = b.id
        if b == 0:
            return self.one()
        elif b < 0:
            return ~self.type(pow(self.id, b * -1, self.P))
        elif b == 1:
            return self
        elif b == (1/2):
            root = modular_sqrt(self.id, self.P)
            assert root != 0, "ins dont have root"
            return self.type(root)
        else:
            return fast_pow(b, self, self.type(1))


class Field(Ring):
    __slots__ = ()

    @abstractmethod
    def sec_inverse(self) -> 'Field':
        '''
        Implement for axiom `inverse`
        '''
        pass

    @classmethod
    def sec_identity(cls):
        return cls(1)

    @classmethod
    def one(cls):
        return cls.sec_identity()

    def __invert__(self):
        return self.sec_inverse()

    def __truediv__(self, g: 'Field') -> 'Field':
        if isinstance(g.id, complex):
            return complex_truediv_algorithm(complex(1), self.id, self.type)
        return self.sec_op(g.sec_inverse())

    def __floordiv__(self, g: 'Field') -> 'Field':
        if isinstance(g.id, complex):
            return complex_truediv_algorithm(complex(1), self.id, self.type)
        return self.sec_op(g.sec_inverse())



# ref: https://blog.decentriq.ch/zk-snarks-primer-part-one/

from klefki.zkp.r1cs import R1CS, mul as vmul
from functools import partial, reduce
from operator import add, mul


def lagrange_polynomial(xs, ys, field=int):
    k = len(xs)

    def l(j): return lambda x: reduce(
        mul, [(x-xs[m])/(xs[j]-xs[m]) for m in range(0, k) if m != j])
    return lambda x: field(reduce(add, [ys[j] * l(j)(x) for j in range(0, k)]))


def transfer(v, field=int):
    ret = []
    for n in range(len(v[0])):
        a = []
        for i in range(len(v)):
            a.append(v[i][n])
        ret.append(lagrange_polynomial([field(i)
                                        for i in range(1, len(a)+1)], a, field))
    return lambda x: [f(x) for f in ret]


def map2field(v, field=int):
    return [[field(j) for j in i]for i in v]


def R1CS2QAP(a, b, c, x=None, field=int):
    #    s = [field(i) for i in s]
    a = map2field(a, field)
    b = map2field(b, field)
    c = map2field(c, field)
    A = transfer(a, field)
    B = transfer(b, field)
    C = transfer(c, field)

    def Z(x): return reduce(mul, [(x - field(i)) for i in range(1, len(a)+1)])
    if not x:
        return A, B, C, Z
    return (A(x), B(x), C(x), Z(x))


def proof(s, A, B, C, Z, field=int):
    s = [field(i) for i in s]
    A = sum(vmul(A, s))
    B = sum(vmul(B, s))
    C = sum(vmul(C, s))
    H = (A * B - C) * (Z ** (-1))
    assert A * B - C == H * Z
    return (s, H)


def verify(s, A, B, C, Z, H):
    return sum(vmul(A, s)) * sum(vmul(B, s)) - sum(vmul(C, s)) == H * Z



import ast
from time import time
from copy import deepcopy


class Flattener:
    def __init__(self, src, ctx={}):
        self.ops = ["set", "+", "-", "*", "/"]
        self.ctx = ctx
        # drop decorator
        src = "\n".join([r for r in src.split("\n") if not "@" in r])
        raw = ast.parse(src.lstrip()).body
        assert len(raw) == 1 and isinstance(
            raw[0], ast.FunctionDef), "only support function"
        self.raw = raw[0]
        self.extra_inputs()
        self.syms = [i for i in self.inputs]
        self.extra_body()
        self._symbol = 0
        self.flatten_body()
        self.closure = {}

    def closure_alias(self, sym, refs):
        if not isinstance(sym, str):
            return sym
        if sym in self.ops:
            return sym
        return "Local<Rc(%s)>%s" % (refs, sym)

    def mk_symbol(self, base="Sym"):
        ret = "%s::%s" % (base, str(self._symbol))
        self._symbol += 1
        self.syms.append(ret)
        return ret

    def latest_sym(self, base, bias=1):
        if not isinstance(base, str):
            return base
        base = base.split("::")[0]
        syms = [s for s in self.syms if s.split("::")[0] == base]
        if len(syms):
            return syms[-bias]
        else:
            return base

    def extra_inputs(self):
        self.inputs = [arg.arg for arg in self.raw.args.args]

    def handle_subscript(self, s, index=None):
        assert isinstance(s, ast.Subscript)
        assert isinstance(s.slice, ast.Index)
        assert isinstance(s.value, ast.Name)
        if isinstance(s.slice.value, ast.Name):
            if index == None:
                index = self.ctx[s.slice.value.id]
        else:
            index = s.slice.value.value
        return ast.Num(self.ctx[s.value.id][index])

    def extra_loop(self, loop):
        """
        only support:
        for _ in range(3):
        """
        avalid_iter_arg = (ast.Constant, ast.Name)
        loop_index = loop.target.id
        assert loop.iter.func.id == "range"
        assert len(loop.iter.args) == 1
        assert isinstance(loop.iter.args[0], avalid_iter_arg)
        if isinstance(loop.iter.args[0], ast.Constant):
            times = loop.iter.args[0].value
        else:
            times = self.ctx[loop.iter.args[0].id]
        ret = [(i, deepcopy(loop.body)) for i in range(times)]
        for e in ret:
            for s in e[1]:
                if isinstance(s, ast.Assign):
                    if isinstance(s.value, ast.Subscript):
                        s.value = self.handle_subscript(s.value, index=e[0])
                    if isinstance(s.value, ast.BinOp):
                        if isinstance(s.value.right, ast.Subscript):
                            s.value.right = self.handle_subscript(
                                s.value.right, index=e[0])
                        if isinstance(s.value.right, ast.Name) and s.value.right.id == loop_index:
                            s.value.right = ast.Num(e[0])
                        if isinstance(s.value.left, ast.Name) and s.value.left.id == loop_index:
                            s.value.left = ast.Num(e[0])
                    if isinstance(s.value, ast.Name) and s.value.id == loop_index:
                        s.value = ast.Num(e[0])

        return sum([r[1] for r in ret], [])

    def extra_body(self):
        body = []
        avalid_stmt = (ast.Assign, ast.Return, ast.For, ast.Assert)
        returned = False
        for c in self.raw.body:
            assert isinstance(c, avalid_stmt)
            assert not returned
            if isinstance(c, ast.For):
                body += self.extra_loop(c)
                continue
            elif isinstance(c, ast.Return):
                returned = True
            body.append(c)
        self.body = body

    def flatten_body(self):
        self.flatten_code = sum([self.flatten_stmt(c) for c in self.body], [])

    def transfer_assert(self, stmt):
        assert isinstance(stmt.test, ast.Compare)
        assert len(stmt.test.ops) == 1
        assert isinstance(stmt.test.ops[0], ast.Eq)
        assert isinstance(stmt.test.left, ast.Name)
        target = stmt.test.left.id
        # dont make new symbol if assert
        # instead, use the latest symbol of target
        target = self.latest_sym(target)
        value = stmt.test.comparators[0]
        return self.flatten_expr(target, value)

    def flatten_stmt(self, s, force_target=None):
        if isinstance(s, ast.Assign):
            assert len(s.targets) == 1 and isinstance(s.targets[0], ast.Name)
            target = s.targets[0].id
            if target in self.syms:
                target = self.mk_symbol(target)
            else:
                self.syms.append(target)
        elif isinstance(s, ast.Assert):
            return self.transfer_assert(s)

        elif isinstance(s, ast.Return):
            target = '~out'
            # if isinstance(s.value, ast.Tuple):
            #     rets = s.value.elts
            #     ret = []
            #     for i in range(len(rets)):
            #         target = self.mk_symbol(target)
            #         self.syms.append(target)
            #         ret.append(
            #             self.flatten_stmt(ast.Return(value=rets[i]), force_target=self.mk_symbol(target))
            #         )
            #     return ret
        # Get inner content
        target = force_target or target
        return self.flatten_expr(target, s.value)

    def flatten_expr(self, target, expr):
        avalid_expr = (ast.Name, ast.Num, ast.BinOp, ast.Call, ast.Constant)

        assert isinstance(expr, avalid_expr), expr
        # x = y
        if isinstance(expr, ast.Name):
            return [['set', target, self.latest_sym(expr.id)]]
        # x = 5
        elif isinstance(expr, ast.Num):
            return [['set', target, expr.n]]
        elif isinstance(expr, ast.Constant):
            return [['set', target, expr.value]]
        elif isinstance(expr, ast.BinOp):
            return self.flatten_binop(target, expr)
        elif isinstance(expr, ast.Call):
            return self.flatten_call(target, expr)

    def flatten_call(self, target, expr):
        fn_ins = self.ctx[expr.func.id]
        if not hasattr(fn_ins, "rc"):
            fn_ins.rc = -1
        fn_ins.rc += 1
        fn_flat = deepcopy(fn_ins.flatcode)
        fn_inputs = deepcopy(fn_ins.inputs)

        # tagging closure vars:
        for f in fn_flat:
            for i in range(len(f)):
                if isinstance(f[i], str):
                    # mark as local rc
                    if not f[i].split("::")[0] in fn_inputs:
                        f[i] = self.closure_alias(f[i], fn_ins.rc)
                    else:
                        # update global vars to latest
                        if f[i] in fn_inputs:
                            if all([target == self.latest_sym(target),
                                    target.split("::")[
                                0] == fn_inputs.index(f[i]),
                                    target != fn_inputs.index(f[i])
                            ]):
                                f[i] = self.latest_sym(
                                    fn_inputs.index(f[i]), 2)
                            else:
                                f[i] = self.latest_sym(
                                    fn_inputs.index(f[i]), 1)
                        else:
                            f[i] = self.closure_alias(f[i], fn_ins.rc)

        fn_flat[-1][1] = target
        return fn_flat

    def flatten_binop(self, target, expr):
        avalid_binop = (ast.Mult, ast.Add, ast.Sub, ast.Div, ast.Pow)
        assert type(expr.op) in avalid_binop
        if isinstance(expr.op, ast.Add):
            op = '+'
        elif isinstance(expr.op, ast.Mult):
            op = '*'
        elif isinstance(expr.op, ast.Sub):
            op = '-'
        elif isinstance(expr.op, ast.Div):
            op = '/'
        elif isinstance(expr.op, ast.Pow):
            return self.flatten_pow(target, expr)

        if isinstance(expr.left, (ast.Name, ast.Num)):
            if isinstance(expr.left, ast.Name):
                # target was shadowed
                if all([target == self.latest_sym(target),
                        target.split("::")[0] == expr.left.id,
                        target != expr.left.id
                        ]):
                    var1 = self.latest_sym(expr.left.id, 2)
                else:
                    var1 = self.latest_sym(expr.left.id, 1)

            sub1 = []
        # If one of the subexpressions is itself a compound expression, recursively
        # apply this method to it using an intermediate variable
        else:
            var1 = self.mk_symbol()
            sub1 = self.flatten_expr(var1, expr.left)
        # Same for right subexpression as for left subexpression
        if isinstance(expr.right, (ast.Name, ast.Num)):
            if isinstance(expr.right, ast.Name):
                if all([target == self.latest_sym(target),
                        target.split("::")[0] == expr.right.id,
                        target != expr.right.id
                        ]):
                    var2 = self.latest_sym(expr.right.id, 2)
                else:
                    var2 = self.latest_sym(expr.right.id, 1)
            else:
                var2 = expr.right.n
            sub2 = []
        else:
            var2 = self.mk_symbol()
            sub2 = self.flatten_expr(var2, expr.right)
        # Last expression represents the assignment; sub1 and sub2 represent the
        # processing for the subexpression if any
        return sub1 + sub2 + [[op, target, var1, var2]]

    def flatten_pow(self, target, expr):
        assert isinstance(expr.right, ast.Num)
        if expr.right.n == 0:
            return [['set', target, 1]]
        elif expr.right.n == 1:
            return flatten_expr(target, expr.left)
        else:  # This could be made more efficient via square-and-multiply but oh well
            if isinstance(expr.left, (ast.Name, ast.Num)):
                if isinstance(expr.left, ast.Name):
                    nxt = base = self.latest_sym(expr.left.id)
                else:
                    expr.left.n
                o = []
            else:
                nxt = base = self.mk_symbol()
                o = self.flatten_expr(base, expr.left)
            for i in range(1, expr.right.n):
                latest = self.latest_sym(nxt)
                nxt = target if i == expr.right.n - 1 else self.mk_symbol()
                o.append(['*', nxt, latest, base])
            return o
